"""Extra functionality for [NumPy](https://numpy.org/)"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/api/utilities/numpy.ipynb.

# %% auto 0
__all__ = ['sym_range']

# %% ../../nbs/api/utilities/numpy.ipynb
import numpy as np
from numpy import array, ndarray

# %% ../../nbs/api/utilities/numpy.ipynb
def sym_range(
    n # Half-length of the array
)->ndarray: # An integer NumPy array running from -n to n
    """A symmetric `arange`."""
    return np.arange(-n, n+1)
